# -*- coding: utf-8 -*-
#
# This file is part of the parce-qt Python package.
#
# Copyright © 2020 by Wilbert Berendsen <info@wilbertberendsen.nl>
#
# This module is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This module is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


"""
This module provides a SyntaxHighlighter.
"""

from PyQt5.QtCore import QEventLoop, QObject, Qt
from PyQt5.QtGui import QGuiApplication, QTextCharFormat, QTextCursor, QTextLayout

import parce.util
import parce.theme

from . import treebuilder
from . import util
from . import formatter


class SyntaxHighlighter(util.SingleInstance):
    """Provides syntax highlighting using parce parsers.

    Instantiate with::

        SyntaxHighlighter.instance(treebuilder)

    You need to set a :class:`~parceqt.formatter.Formatter` which supplies
    the highlighting text formats.

    By default, no formatter is set; use ``set_formatter()`` to set a
    Formatter, which is needed to enable highlighting.

    """
    def __init__(self, builder):
        self._formatter = None
        self._cursor = None      # remembers the range to rehighlight
        builder.updated.connect(self.slot_updated)
        builder.preview.connect(self.slot_preview, Qt.BlockingQueuedConnection)

    def builder(self):
        """Return the builder we were instantiated with."""
        return self.target_object()

    def delete(self):
        """Reimplemented to clear the highlighting before delete."""
        self.clear()
        super().delete()

    def set_formatter(self, formatter):
        """Set the Formatter to use, or None for no formatter.

        The Formatter provides the text formats to highlight.
        If you set the Formatter to None, highlighting is effectively disabled,
        although the TreeBuilder still does the tokenizing.

        """
        if formatter is not self._formatter:
            self._formatter = formatter
            self.rehighlight()

    def formatter(self):
        """Return the currently set Formatter."""
        return self._formatter

    def clear(self):
        """Clear the highlighting.

        This method is called when the formatter is set to None, and when
        SyntaxHighlighter is explicitedly :meth`delete`-d.

        """
        doc = self.document()
        block = doc.firstBlock()
        while block.isValid():
            block.layout().clearFormats()
            block = block.next()
        doc.markContentsDirty(0, doc.characterCount() - 1)

    def rehighlight(self):
        """Draw or clear the highlighting, depending on the Formatter.

        This method is automatically called when the Formatter
        is changed. Clears the highlighting if the current formatter is None.

        """
        if self._formatter:
            root = self.builder().get_root()
            if root:
                # no need to redraw if treebuilder is already busy
                end = self.document().characterCount() - 1
                self.draw_highlighting(root, 0, end, True)
        else:
            self.clear()

    def document(self):
        """Return the QTextDocument."""
        return self.builder().document()

    def slot_preview(self, start, tree):
        """Called when there is a peek preview."""
        if self._formatter:
            doc = self.document()
            end = doc.findBlock(tree.end).position() - 1
            if start < end:
                self.draw_highlighting(tree, start, end)

    def slot_updated(self, start, end):
        """Called on update; performs the highlighting."""
        if self._formatter:
            self.draw_highlighting(self.builder().root, start, end, True)

    def draw_highlighting(self, root, start, end, interruptible=False):
        """Draw the highlighting using tree ``root`` from ``start`` to ``end``.

        If ``interruptible`` is True, ``QApplication::process_events()`` is
        called every 1000 lines, to enable the user typing in the document,
        which also causes the highlighting to quit and resume later.

        """
        if not self._formatter:
            return

        doc = self.document()

        if interruptible:
            c = self._cursor
            if c:
                # our previous highlighting run was interrupted, fix it now
                start = min(start, c.selectionStart())
                end = max(end, c.selectionEnd())
            else:
                c = self._cursor = QTextCursor(doc)
                c.setKeepPositionOnInsert(True)
        else:
            c = None

        block = doc.findBlock(start)
        pos = block.position()
        last_block = doc.findBlock(end)
        if not last_block.isValid():
            last_block = doc.lastBlock()
        end = last_block.position() + last_block.length() - 1

        if c:
            c.setPosition(end)

        num = block.blockNumber() + 100
        formats = split_formats(block, start)[0]
        for f in self._formatter.format_ranges(root, start, end):
            while f.pos >= pos + block.length():
                block.layout().setFormats(formats)
                block = block.next()
                pos = block.position()
                formats = []
            r = QTextLayout.FormatRange()
            r.format = f.textformat
            r.start = f.pos - pos
            f_end = min(end, f.end)
            while f_end > pos + block.length():
                r.length = block.length() - r.start - 1
                formats.append(r)
                block.layout().setFormats(formats)
                block = block.next()
                pos = block.position()
                formats = []
                r = QTextLayout.FormatRange()
                r.format = f.textformat
                r.start = 0
            r.length = f_end - pos - r.start
            formats.append(r)
            if c and block.blockNumber() > num:
                num = block.blockNumber() + 1000
                doc.markContentsDirty(start, pos - start)
                start = pos
                c.setPosition(start, QTextCursor.KeepAnchor)
                revision = doc.revision()
                QGuiApplication.processEvents(QEventLoop.ExcludeSocketNotifiers)
                # if the user typed, immediately quit, but come back!
                if doc.revision() != revision:
                    return
        block.layout().setFormats(formats)
        while block < last_block:
            block = block.next()
            block.layout().clearFormats()
        doc.markContentsDirty(start, end - start)
        # we have finished highlighting
        if c:
            self._cursor = None


def split_formats(block, position):
    """Return two lists of FormatRange instances from the block's layout.

    The first are the formats from the start of the block until position (which
    must be inside the block). The second are the formats from the position
    to the end of the block, shifted as if the block started at position.

    A format range is neatly cut in two when position lies in the middle of
    a range.

    """
    def new_format_range(r, offset=0):
        n = QTextLayout.FormatRange()
        n.format = r.format
        n.start = r.start + offset
        n.length = r.length
        return n
    pos = position - block.position()
    formats = iter(block.layout().formats())
    start_formats = []
    end_formats = []
    for r in formats:
        if r.start < pos:
            n = new_format_range(r)
            start_formats.append(n)
            if r.start + r.length > pos:
                n.length = pos - r.start
                n = new_format_range(r)
                n.start = 0
                n.length = r.start + r.length - pos
                end_formats.append(n)
                break
        else:
            n = new_format_range(r, -pos)
            end_formats.append(n)
            break
    for r in formats:
        n = new_format_range(r, -pos)
        end_formats.append(n)
    return start_formats, end_formats





# -*- coding: utf-8 -*-
#
# This file is part of the parce-qt Python package.
#
# Copyright © 2020 by Wilbert Berendsen <info@wilbertberendsen.nl>
#
# This module is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This module is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


"""
This module implements a Document encapsulating a QTextDocument.

It is not needed to store the Document itself, it is only used
to modify the QTextDocument through the parce.Document API.

We do not ourself retokenize the text, that is done by a TreeBuilder
that is automatically connected to the document.

"""

from PyQt5.QtGui import QTextCursor

import parce.treedocument
import parce.document

from . import treebuilder


class Document(
    parce.treedocument.TreeDocumentMixin, parce.document.AbstractDocument):
    """Document accesses a QTextDocument via the parce.Document API.

    There is no need to store this object, it is only used to access and
    modify the contents of a QTextDocument. Example::

        d = Document(doc)   # where doc is an existing QTextDocument
        with d:
            d[5:5] = 'some text'

    This is useful when you have written code that manipulates text files based
    on the tokenized tree via the parce.Document API, you can use the same code
    to manipulate QTextDocuments in an interactive session.

    As opposed to parce.Document, this Document class does not update the token
    tree by itself, that is handled by the TreeBuilder that does its work in
    the background.

    """
    def __init__(self, document):
        """Initialize with QTextDocument."""
        parce.document.AbstractDocument.__init__(self)
        self._document = document

    def builder(self):
        """Reimplemented to return the builder that's active on our QTextDocument."""
        return treebuilder.TreeBuilder.instance(self.document())

    def document(self):
        """Return our QTextDocument."""
        return self._document

    def text(self):
        """Reimplemented to get the text from the QTextDocument."""
        return self.document().toPlainText()

    def __len__(self):
        """Reimplemented to return the length of the text in the QTextDocument."""
        # see https://bugreports.qt.io/browse/QTBUG-4841
        return self.document().characterCount() - 1

    def _update_contents(self):
        """Apply the changes to our QTextDocument."""
        doc = self.document()
        c = QTextCursor(self.document())
        c.beginEditBlock()
        for start, end, text in reversed(self._changes):
            c.setPosition(end)
            if start != end:
                c.setPosition(start, QTextCursor.KeepAnchor)
            c.insertText(text)
        c.endEditBlock()

    def _get_contents(self, start, end):
        """Reimplemented to get a fragment of our text.

        This is faster than getting the whole text and using Python to slice it.

        """
        doc = self.document()
        c = QTextCursor(self.document())
        c.setPosition(end)
        c.setPosition(start, QTextCursor.KeepAnchor)
        return c.selection().toPlainText()

    def contents_changed(self, start, removed, added):
        """Reimplemented to do nothing, it is already handled by TreeBuilder."""
        pass


class Cursor(parce.document.Cursor):
    """A cursor with a textCursor() method to return a QTextCursor.

    Only use this Cursor with parceqt.Document.

    """
    def textCursor(self):
        c = QTextCursor(self.document().document())
        if self.end is None:
            c.movePosition(QTextCursor.End)
        else:
            c.setPosition(self.end)
        c.setPosition(self.start, QTextCursor.KeepAnchor)
        return c





# -*- coding: utf-8 -*-
#
# This file is part of the parce-qt Python package.
#
# Copyright © 2020 by Wilbert Berendsen <info@wilbertberendsen.nl>
#
# This module is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This module is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


"""
Various utility classes and functions.
"""

import weakref

from PyQt5.QtCore import QThread


_jobs = set()   # store running Job instances


class Job(QThread):
    """Simple QThread helper that runs a ``function`` in a background thread.

    If you specify a ``finished`` callable, it will be called when the function
    has finished. If ``with_result`` is True, the ``finished`` callable will
    be called with the result of the function.

    The job is started immediately. You do not need to store the job, it will
    keep a reference itself as long as it is running.

    """
    def __init__(self, function, finished=None, with_result=False):
        super().__init__()
        self._function = function
        self._finished = finished
        self._with_result = with_result
        _jobs.add(self)
        self.finished.connect(self._slot_finished)
        self.start()

    def run(self):
        """Run the function and store the result."""
        self._result = self._function()

    def _slot_finished(self):
        _jobs.discard(self)
        if self._finished is not None:
            if self._with_result:
                self._finished(self._result)
            else:
                self._finished()


class SingleInstance:
    """Keeps a single instance around for another object.

    We keep only a weak reference to the object, if the object is garbage
    collected, we disappear as well.

    """
    @classmethod
    def instance(cls, obj, *args, **kwargs):
        """Get or create the instance for obj."""
        try:
            return cls._instances[obj]
        except AttributeError:
            cls._instances = weakref.WeakKeyDictionary()
        except KeyError:
            pass
        # target must be in place before init is called
        new = cls._instances[obj] = cls.__new__(cls)
        new._target = weakref.ref(obj)
        new.__init__(obj, *args, **kwargs)
        return new

    @classmethod
    def delete_instance(cls, obj):
        """Actively remove the stored instance."""
        try:
            instance = cls._instances[obj]
        except (KeyError, AttributeError):
            pass
        else:
            instance.delete()

    @classmethod
    def get_instance(cls, obj):
        """Return the instance if it already exists, else returns None."""
        try:
            return cls._instances[obj]
        except (AttributeError, KeyError):
            pass

    def target_object(self):
        """Return the object we are stored for."""
        return self._target()

    def delete(self):
        """Delete the stored reference to ourself.

        Other cleanup code could be added by reimplementing this method.
        This is not called on garbage collection, but only when called
        directly or via delete_instance().

        """
        obj = self._target()
        if obj is not None:
            del self.__class__._instances[obj]


class Switch:
    """A context manager that evaluates to True when in a context, else to False.

    Example::

        clicking = Switch()

        def myfunc():
            with clicking:
                blablabl()

        # and elsewhere:
        def blablabl():
            if not clicking:
                do_something()

        # when blablabl() is called from myfunc, clicking evaluates to True,
        # so do_something() is not called then.

    """
    def __init__(self):
        self._value = 0

    def __enter__(self):
        self._value += 1

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._value -= 1

    def __bool__(self):
        return bool(self._value)


def call_async(function, finished=None):
    """Call ``function()`` in a background thread and then ``finished()`` when
    done in the main thread."""
    return Job(function, finished)


def call_async_with_result(function, finished=None):
    """Call ``result = function()`` in a background thread and then
    ``finished(result)`` when done in the main thread."""
    return Job(function, finished, True)





# -*- coding: utf-8 -*-
#
# This file is part of the parce-qt Python package.
#
# Copyright © 2020 by Wilbert Berendsen <info@wilbertberendsen.nl>
#
# This module is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This module is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


"""
The Python parceqt module provides parce parsing and highlighting
features for Qt's QTextDocument.

This module depends on the parce module.

With a few simple function calls you can highlight the syntax of
QTextDocument using parce.

"""

__all__ = (
    'Cursor',
    'Document',
    'Formatter',
    'TreeBuilder',
    'SyntaxHighlighter',
    'builder',
    'root',
    'set_root_lexicon',
    'root_lexicon',
    'highlight',
    'adjust_widget',
    'cursor',
    'version',
    'version_string',
)

from PyQt5.QtWidgets import QApplication

import parce.theme

from .pkginfo import version, version_string
from .document import Cursor, Document
from .formatter import Formatter
from .treebuilder import TreeBuilder
from .highlighter import SyntaxHighlighter


def builder(doc):
    """Return the TreeBuilder responsible for tokenizing this QTextDocument.

    If no TreeBuilder already existed, it is instantiated and becomes a child
    of the QTextDocument. You can connect to its ``updated()`` or ``changed()``
    signal to get notified of changes in the tokenized tree.

    """
    return TreeBuilder.instance(doc)


def root(doc, wait=False):
    """Get the root element of the tokenized tree of specified text document.

    See for more information about the arguments the ``get_root()`` method
    of ``treebuilder.TreeBuilder``.

    """
    return builder(doc).get_root(wait)


def set_root_lexicon(doc, lexicon):
    """Instatiate a TreeBuilder for the document if needed, and set its root lexicon."""
    builder(doc).set_root_lexicon(lexicon)


def root_lexicon(doc):
    """Return the currently active root lexicon for the QTextDocument."""
    return builder(doc).root_lexicon()


def highlight(doc, theme="default"):
    """Set the highlighting Theme for the document.

    Use a string value to select a Theme by name. Use None to disable
    highlighting, or use False to force the SyntaxHighlighter to quit.

    Of course, highlighting becomes only visible when the document has a
    root_lexicon set.

    """
    b = builder(doc)
    if theme is False:
        SyntaxHighlighter.delete_instance(b)
    else:
        if isinstance(theme, str):
            theme = parce.theme.Theme.byname(theme)
        formatter = Formatter(theme) if theme else None
        SyntaxHighlighter.instance(b).set_formatter(formatter)


def adjust_widget(widget):
    """Convenience function to set palette and font of a text editing ``widget``.

    Sets the widget's palette and font to the theme of its QTextDocument's
    highlighter.

    The widget must be a QPlainTextEdit, QTextEdit or QTextBrowser. If its
    document has not yet a theme set, this function does nothing.

    Basically this is as simple as::

        formatter = parceqt.formatter.Formatter(theme)
        widget.setFont(formatter.font())
        widget.setPalette(formatter.palette())

    but this function is useful when you just set the theme once to a document
    and want to have its editing widget adjusted.

    Also, when you stopped the highlighting, this function neatly switches the
    widget back to the default palette and font.

    """
    doc = widget.document()
    b = builder(doc)
    source = QApplication
    h = SyntaxHighlighter.get_instance(b)
    if h:
        source = h.formatter()
    widget.setFont(source.font(widget))
    widget.setPalette(source.palette(widget))


def cursor(cur):
    """Convenience function to return a Cursor for a Document that wraps a QTextDocument.

    You can alter the document via the parce.Document API.
    The returned Cursor has a textCursor() method that returns a QTextCursor
    for the same selection or position.

    """
    c = Cursor(Document(cur.document()))
    c.start = cur.selectionStart()
    c.end = cur.selectionEnd()
    return c


